from vulnix.nvd import NVD, Archive, decompress
from vulnix.whitelist import Whitelist
import pkg_resources
import pytest


@pytest.fixture
def nvd_modified(tmpdir):
    nvd = NVD(cache_dir=str(tmpdir))
    a = Archive('Modified')
    with open(pkg_resources.resource_filename(
            'vulnix', 'tests/nvdcve-2.0-Modified.xml.gz'), 'rb') as f:
        a.parse(decompress(f, str(tmpdir)))
    nvd.add(a)
    return nvd


@pytest.fixture
def whitelist_toml():
    return pkg_resources.resource_stream(
        'vulnix', 'tests/fixtures/whitelist.toml')


@pytest.fixture
def whitelist_yaml():
    return pkg_resources.resource_stream(
        'vulnix', 'tests/fixtures/whitelist.yaml')


@pytest.fixture
def whitelist():
    return Whitelist.load(pkg_resources.resource_stream(
        'vulnix', 'tests/fixtures/whitelist.toml'))



from .utils import batch
from BTrees import OOBTree
from lxml import etree
from persistent import Persistent
import ZODB
import ZODB.FileStorage
import datetime
import gzip
import logging
import os
import os.path as p
import requests
import shutil
import tempfile
import time
import transaction

DEFAULT_MIRROR = 'https://static.nvd.nist.gov/feeds/xml/cve/'
DEFAULT_CACHE_DIR = '~/.cache/vulnix'

NS = {'feed': 'http://scap.nist.gov/schema/feed/vulnerability/2.0',
      'vuln': 'http://scap.nist.gov/schema/vulnerability/0.4'}

logger = logging.getLogger(__name__)


class NVD(object):
    """Access to the National Vulnerability Database.

    https://nvd.nist.gov/
    """

    has_updates = False

    def __init__(self, mirror=DEFAULT_MIRROR, cache_dir=DEFAULT_CACHE_DIR):
        self.mirror = mirror.rstrip('/') + '/'
        self.cache_dir = p.expanduser(cache_dir)
        self._root = {}
        self._root['meta'] = Meta()
        self._root['archives'] = {}

        current_year = datetime.date.today().year
        self.relevant_archives = [
            str(x) for x in range(current_year - 5, current_year + 1)]
        self.relevant_archives.append('Modified')

    def __enter__(self):
        if self._root['archives'].keys():
            raise RuntimeError(
                'Either use an in-memory NVD database or the ZODB backed '
                'variant - not both!',
                'Keys present', self._root.keys())
        logger.debug('Using cache in %s', self.cache_dir)
        if not p.exists(self.cache_dir):
            os.makedirs(self.cache_dir)
        storage = ZODB.FileStorage.FileStorage(
            p.join(self.cache_dir, 'Data.fs'))
        self._db = ZODB.DB(storage)
        self._connection = self._db.open()
        self._root = self._connection.root()
        self._root.setdefault('meta', Meta())
        self._root.setdefault('archives', OOBTree.OOBTree())

    def __exit__(self, exc_type, exc_value, exc_tb):
        if exc_type is None:
            transaction.commit()
            meta = self._root['meta']
            if self.has_updates:
                meta.unpacked += 1
                if meta.unpacked > 25:
                    logger.debug('Packing database')
                    self._db.pack()
                    meta.unpacked = 0
                transaction.commit()
        else:
            transaction.abort()
        self._connection.close()

    def by_product_name(self, name):
        for archive in self._root['archives'].values():
            yield from archive.by_product_name(name)

    def add(self, archive):
        """Inserts an Archive object."""
        if archive.name not in self._root['archives']:
            self._root['archives'][archive.name] = archive

    def update(self):
        # Add missing archives
        for a in self.relevant_archives:
            self.add(Archive(a))

        # Remove superfluous archives
        for a in self._root['archives']:
            if a not in self.relevant_archives:
                del self._root['archives'][a]

        for archive in self._root['archives'].values():
            self.has_updates |= archive.update(self.mirror)
        transaction.commit()


class Meta(Persistent):
    """A small grab bag to store persistent meta information in
    the database.

    """

    unpacked = 0


def decompress(fileobj, dir=None):
    """Decompresses gzipped XML data into a temporary file.

    Returns temporary file. The callee takes responsibility to remove
    that file after use.
    """
    tf = tempfile.NamedTemporaryFile(
        'wb', prefix='vulnix.nvd.', suffix='.xml', delete=False, dir=dir)
    logger.debug("Uncompressing {}".format(tf.name))
    with gzip.open(fileobj, 'rb') as f_in:
        shutil.copyfileobj(f_in, tf)
    tf.close()
    return tf.name


class Download:
    """Wrapper for downloading compressed XML data.

    Uncompressed XML is written to a temporary file and deleted once the
    context is exited.
    """

    def __init__(self, url):
        self.url = url
        self.xml = None

    def __enter__(self):
        logger.debug("Downloading {}".format(self.url))
        r = requests.get(self.url, stream=True, timeout=300)
        r.raise_for_status()
        self.xml = decompress(r.raw)
        return self.xml

    def __exit__(self, exc_type, exc_value, exc_tb):
        os.unlink(self.xml)
        self.xml = None
        return False  # re-raise


class Archive(Persistent):

    last_update = 0

    def __init__(self, name):
        self.name = name
        self.products = OOBTree.OOBTree()
        # Either set to a duration to update every `age_limit` seconds or to
        # None to never update after the initial fetch.
        if name == 'Modified':
            # Is updated every two hours.
            self.age_limit = 3600
        elif name == str(datetime.date.today().year):
            # The current year is only updated every 8 days
            # (folding in the data from Modified).
            self.age_limit = 4 * 86400
        else:
            # Check for errata (quite infrequent).
            self.age_limit = 30 * 86400

    @property
    def upstream_filename(self):
        return 'nvdcve-2.0-{}.xml.gz'.format(self.name)

    @property
    def is_current(self):
        if self.age_limit is None:
            # We don't want to update and we have been fetched before - we're
            # good.
            return bool(self.last_update)
        age = time.time() - self.last_update
        return age < self.age_limit

    def by_product_name(self, name):
        return self.products.get(name, [])

    def update(self, mirror):
        if self.is_current:
            logger.debug('"{}" is up-to-date'.format(self.name))
            return False
        self.products.clear()
        logger.info('Updating "{}"'.format(self.name))
        with Download(mirror + self.upstream_filename) as xml:
            self.parse(xml)
        self.last_update = time.time()
        return True

    def parse(self, filename):
        logger.debug("Parsing {}".format(filename))
        parser = etree.iterparse(
            filename, tag='{' + NS['feed'] + '}entry')
        for event, node in batch(parser, 500, transaction.savepoint):
            vx = Vulnerability.from_node(node)
            # We don't use a ZODB set here as we a) won't ever change this
            # again in the future (we just rebuild the tree) and also I want to
            # avoid making millions of micro-records.
            for cpe in vx.affected_products:
                self.products.setdefault(cpe.product, set())
                self.products[cpe.product].add(vx)
            # We need to explicitly clear this node. iterparse only builds the
            # tree incrementally but does not remove data that isn't needed any
            # longer.  See
            # http://www.ibm.com/developerworks/xml/library/x-hiperfparse/
            node.clear()
            while node.getprevious() is not None:
                del node.getparent()[0]


class Vulnerability(Persistent):

    cve_id = None
    affected_products = ()

    def __init__(self):
        self.affected_products = []

    def __repr__(self):
        return '<Vulnerability({}, {} affected)>'.format(
            self.cve_id, len(self.affected_products))

    @classmethod
    def from_node(cls, node):
        self = cls()
        self.cve_id = node.get('id')
        affected_products = {}
        for product in node.findall('.//vuln:product', NS):
            cpe = CPE.from_uri(product.text)
            if cpe.product is None:
                # This usually indicates a combination where some vulnerability
                # only applies for a specific operating system *vendor*
                continue
            key = (cpe.vendor, cpe.product)
            master_cpe = affected_products.setdefault(key, cpe)
            master_cpe.versions.update(cpe.versions)
        self.affected_products = list(affected_products.values())
        return self


class CPE(Persistent):

    # These are the only attributes we're interested in. Reduce memory
    # footprint by not storing unused attributes.
    vendor = None
    product = None
    versions = None

    @classmethod
    def from_uri(cls, uri):
        self = cls()
        self.versions = set()
        protocol, identifier = uri.split(':/')
        assert protocol == 'cpe'
        component_list = identifier.split(':')
        components = ['part', 'vendor', 'product', 'version', 'update',
                      'edition', 'lang']
        while component_list:
            component_name = components.pop(0)
            component_value = component_list.pop(0)
            if component_name == 'version':
                self.versions.add(component_value)
            elif hasattr(self, component_name):
                setattr(self, component_name, component_value)
        return self

    def __repr__(self):
        return '<CPE({}:{})>'.format(self.product, repr(self.versions))



from .derivation import load, split_name, NoVersionError
from .utils import call

import logging
import os.path as p

_log = logging.getLogger(__name__)


class Store(object):

    def __init__(self, requisites=True):
        self.requisites = requisites
        self.derivations = {}

    def add_gc_roots(self):
        """Add derivations found for all live GC roots.

        Note that this usually includes old system versions.
        """
        _log.debug('loading all live derivations')
        for d in call(['nix-store', '--gc', '--print-live']).splitlines():
            self.update(d)

    def add_path(self, path):
        """Add the closure of all derivations referenced by a store path."""
        if not p.exists(path):
            raise RuntimeError('path `{}` does not exist - cannot load '
                               'derivations referenced from it'.format(path))
        _log.debug('loading derivations referenced by "%s"', path)
        if path.endswith('.drv'):
            deriver = path
        else:
            deriver = call(['nix-store', '-qd', path]).strip()
            _log.debug('deriver: %s', deriver)
            if not deriver or deriver == 'unknown-deriver':
                raise RuntimeError(
                    'Cannot determine deriver. Is this really a path into the '
                    'nix store?', path)
        if self.requisites:
            for candidate in call(['nix-store', '-qR', deriver]).splitlines():
                self.update(candidate)
        else:
            self.update(deriver)

    def update(self, drv_path):
        if not drv_path.endswith('.drv'):
            return
        if drv_path in self.derivations:
            return
        # quick'n'dirty check if this is a derivation with version
        # load() may fail even if this pre-check succeeds as the derivation's
        # 'name' attribute is the final source of truth
        _pname, version = split_name(p.basename(drv_path)[:-4])
        if not version:
            return
        _log.debug('loading %s', drv_path)
        try:
            drv_obj = load(drv_path)
        except NoVersionError:
            return
        self.derivations[drv_path] = drv_obj



from vulnix.resource import Resource, open_resources
import http.server
import os
import pkg_resources
import pytest
import signal

# local file prefix to the fixtures directory
local = pkg_resources.resource_filename('vulnix', 'tests/fixtures/')


@pytest.yield_fixture
def http_server():
    """Spawns a HTTP server in a separate process, serving test fixtures.

    Yields base URL of the HTTP server (e.g., http://localhost:1234/)
    """
    oldcwd = os.getcwd()
    os.chdir(pkg_resources.resource_filename('vulnix', 'tests/fixtures'))
    httpd = http.server.HTTPServer(
        ('localhost', 0), http.server.SimpleHTTPRequestHandler)
    port = httpd.server_port
    child = os.fork()
    if child == 0:
        signal.alarm(3600)  # safety belt
        httpd.serve_forever()
        return  # never reached
    os.chdir(oldcwd)
    yield 'http://localhost:{}/'.format(port)
    os.kill(child, signal.SIGTERM)
    os.wait()


def test_open_local():
    fn = local + 'whitelist.toml'
    with Resource(fn).open() as f:
        assert f.read() == open(fn, 'rb').read()


def test_open_remote(http_server):
    with Resource(http_server + 'whitelist.toml').open() as f:
        assert f.read() == open(local + 'whitelist.toml', 'rb').read()


def test_multiple_resources(http_server):
    expected = open(local + 'cpio-2.12.drv', 'rb').read()
    gen = open_resources(sources=[
        local + 'cpio-2.12.drv',
        local + 'no-such-file',
        http_server + 'cpio-2.12.drv',
        http_server + 'file-not-found',
    ])
    assert next(gen).read() == expected  # local file
    assert next(gen).read() == expected  # remote
    with pytest.raises(StopIteration):
        next(gen)  # should skip nonexistent files/urls silently



import contextlib
import logging
import urllib
import re

_log = logging.getLogger(__name__)


class Resource:

    R_URL_LIKE = re.compile(r'^[a-z]+://')

    def __init__(self, source, timeout=60):
        self.source = source
        self.remote = self.R_URL_LIKE.match(source)
        self.timeout = timeout

    @contextlib.contextmanager
    def open(self):
        if self.remote:
            yield urllib.request.urlopen(
                self.source, timeout=self.timeout)
        else:
            yield open(self.source, 'rb')


def open_resources(_click_ctx=None, _click_param=None, sources=None):
    """Yields read-only binary fobjs for all given sources.

    Resources with open/connection errors are ignored and warnings are
    logged.

    Designed for use as click option callback.
    """
    if sources:
        for s in sources:
            try:
                with Resource(s).open() as f:
                    yield f
            except (EnvironmentError, urllib.error.URLError) as e:
                _log.warning('failed to open %s: %s', s, e)



#!/usr/bin/env python3
"""Takes JSON output and creates a vulnerability roundup issue.

Feed the output of `vulnix --json` into this script to create the issue
description.
"""

import urllib.parse
import argparse
import json
import sys


def whitelist(pkgs, ignore):
    res = []
    for pkg in pkgs:
        affected_by = [cve for cve in pkg['affected_by'] if cve not in ignore]
        pkg['affected_by'] = affected_by
        if len(affected_by):
            res.append(pkg)
    return res


def parse(json_file):
    if json_file:
        return json.load(json_file)
    return json.load(sys.stdin)


def fmt(pkgs):
    for pkg in sorted(pkgs, key=lambda p: p['name']):
        name = pkg['name']
        quotedname = urllib.parse.quote_plus(pkg['pname'])
        search = 'https://search.nix.gsc.io/?q={}&i=fosho&repos=nixos-nixpkgs'.\
            format(quotedname)
        files = 'https://github.com/NixOS/nixpkgs/search?utf8=%E2%9C%93&q={}+in%3Apath&type=Code'.\
            format(quotedname)
        print("### {name} ([search]({search}), [files]({files}))".\
              format(name=name, search=search, files=files))
        cves = ("[{}]({})".format(
            cve, "https://web.nvd.nist.gov/view/vuln/detail?vulnId="+cve)
            for cve in sorted(pkg['affected_by']))
        for cve in cves:
            print(' - [ ] {}'.format(cve))
        print()

def main():
    a = argparse.ArgumentParser(description=__doc__)
    a.add_argument('VULNIX_JSON', nargs='?', type=argparse.FileType('r'),
                   help='vulnix json output (file or "-")')
    a.add_argument('-i', '--ignore', type=argparse.FileType('r'),
                   help='List of CVEs to ignore (one per line)')
    a.add_argument('-g', '--git-id', default='HEAD',
                   help='nixpkgs commit used for scanning')
    args = a.parse_args()
    if args.ignore:
        ignore = set(args.ignore.read().split())
    else:
        ignore = set()
    print('Scanned nixos/release-combined.nix @ {id}. Filtered out '
          'previously reported CVEs. May contain false positives.\n'.
          format(id=args.git_id[:7]))
    fmt(whitelist(parse(args.VULNIX_JSON), ignore))
    print("""\
Cc: @joepie91, @phanimahesh, @the-kenny, @7c6f434c, @k0001, @peterhoeg, @nh2, @LnL7, @grahamc, @adisbladis, @fpletz

Contact @ckauhaus for any questions.
""")


if __name__ == '__main__':
    main()



# This should be only one line. If it must be multi-line, indent the second
# line onwards to keep the PKG-INFO file format intact.
"""Scans a Nix store for derivations that are affected by vulnerabilities."""

from setuptools import setup, find_packages
import os.path


def project_path(*names):
    return os.path.join(os.path.dirname(__file__), *names)

with open(project_path('VERSION')) as f:
    version = f.read().strip()

long_description = []

for rst in ['README.rst', 'HACKING.rst', 'CHANGES.rst']:
    with open(project_path(rst)) as f:
        long_description.append(f.read())

setup(
    name='vulnix',
    version=version,
    install_requires=[
        'click==6.7',
        'colorama==0.3.9',
        'lxml==3.8.0',
        'pyyaml==3.12',
        'requests==2.18.3',
        'toml>=0.9',
        'ZODB==5.2.4',
    ],
    extras_require={
        'test': [
            'freezegun>0.3',
            'pytest>=3.2',
            'pytest-cov>=2.5',
            'pytest-flake8',
            'pytest-runner>=2.11,<3dev',
            'pytest-timeout>=1.2',
            'setuptools_scm>=1.15',
        ],
    },
    entry_points="""
        [console_scripts]
            vulnix = vulnix.main:main
    """,
    author='Flying Circus Internet Operations GmbH',
    author_email='mail@flyingcircus.io',
    license='BSD-3-Clause',
    url='https://github.com/flyingcircusio/vulnix',
    keywords='security',
    classifiers="""\
Development Status :: 5 - Production/Stable
Environment :: Console
Intended Audience :: System Administrators
License :: OSI Approved :: BSD License
Operating System :: POSIX
Programming Language :: Python
Programming Language :: Python :: 3 :: Only
Programming Language :: Python :: 3.4
Programming Language :: Python :: 3.5
Programming Language :: Python :: 3.6
Topic :: System :: Systems Administration
"""[:-1].split('\n'),
    description=__doc__.strip(),
    long_description='\n\n'.join(long_description),
    packages=find_packages('src'),
    package_dir={'': 'src'},
    include_package_data=True,
    zip_safe=False
)
